
#include "quakedef.h"
#include "cl_collision.h"

/*
// not yet used
typedef struct physentity_s
{
	// this may be a entity_t, or a prvm_edict_t, or whatever
	void *realentity;

	// can be NULL if it is a bbox object
	model_t *bmodel;

	// node this entity crosses
	// for avoiding unnecessary collisions
	physnode_t *node;

	// matrix for converting from model to world coordinates
	double modeltoworldmatrix[3][4];

	// matrix for converting from world to model coordinates
	double worldtomodelmatrix[3][4];

	// if this is a bmodel, this is used for culling it quickly
	// if this is not a bmodel, this is used for actual collisions
	double mins[3], maxs[3];
}
physentity_t;
*/

// VENOMUS
void CL_GetTraceTextureAndUVCoords (trace_t *trace, qboolean autoGetUV)
{
	msurface_t *surface, *endsurface;
	int i, endtriangle;
	int *tri;
	float *v1, *v2, *v3, *tc1, *tc2, *tc3, *min, *max, *pmin, *pmax;
	vec3_t l1, l2, x1, x2, x3;
	vec3_t pminToPmax, pminToEndpos;
	vec3_t traceEndpos;
	vec3_t traceEndposLocal;

// moving bmodel fix
	VectorCopy (trace->endpos, traceEndpos);
	Matrix4x4_Transform (&cl.entities[trace->hitEnt].render.inversematrix, traceEndpos, traceEndposLocal);

	if ((!trace->getTex && !trace->getUV) || trace->fraction >= 1)
	{
		return;
	}
// The clipping hull 0 is a direct copy of the visible hull, therefore we should be able to find the node
// (flat surface) of the visible that has the same node number as the hit clipping node of hull 0 on the trace.
// But the node is composed of several smaller areas known as surfaces, that may have different textures and/or
// texture coordinates. Therefore we need to find the actual surface we hit. We do this by checking if the endpoint
// is 'on' each triangle of every surface in the node (surfaces are further divided into triangular polygons that
// form the drawing mesh).
// first loop through all surfaces of the node
	for
	(
		surface = trace->model->data_surfaces + trace->model->brush.data_nodes[trace->nodenum].firstsurface,
		endsurface = surface + trace->model->brush.data_nodes[trace->nodenum].numsurfaces;
		surface < endsurface;
		surface ++
	)
	{
// the surface must be a solid wall surface
		if (surface->texture->supercontents & SUPERCONTENTS_SOLID)
		{
// Loop through all the triangles of each surface. The triangle, vertex and texture coordinate data are actually
// stored as big arrays for every triangle of every surface of every node in a brush model. But the surface structs
// provide offsets to their own section of the mesh data.
			endtriangle = (surface->num_firsttriangle + surface->num_triangles) * 3;
			for (i = surface->num_firsttriangle * 3; i < endtriangle; i += 3)
			{
				tri = trace->model->surfmesh.data_element3i + i;
				v1 = trace->model->surfmesh.data_vertex3f + *(tri + 0) * 3;
				v2 = trace->model->surfmesh.data_vertex3f + *(tri + 1) * 3;
				v3 = trace->model->surfmesh.data_vertex3f + *(tri + 2) * 3;

				VectorSubtract (v1, v2, l1);
				VectorSubtract (v1, traceEndposLocal, l2);
				CrossProduct (l1, l2, x1);

				VectorSubtract (v2, v3, l1);
				VectorSubtract (v2, traceEndposLocal, l2);
				CrossProduct (l1, l2, x2);

				VectorSubtract (v3, v1, l1);
				VectorSubtract (v3, traceEndposLocal, l2);
				CrossProduct (l1, l2, x3);

// If these three dotproducts > 0, traceEndposLocal is 'on the triangle', so we know we have found the right surface (and can
// also work out texture coordinates from the vertices and uv data).
				if (DotProduct (x1, x2) > 0)
				{
					if (DotProduct (x2, x3) > 0)
					{
						if (DotProduct (x3, x1) > 0)
						{
							trace->tex = surface->texture;
							
							if (trace->tex)
							{
								if (autoGetUV && trace->tex->material->colorMode == IMPACT_COLOR_MODE_PERPIXEL)
								{
									trace->getUV = TRUE;
								}
// do simple linear interpolation for the traceEndposLocal 'texture coordinates'
								if (trace->getUV)
								{
// first for u...
									tc1 = trace->model->surfmesh.data_texcoordtexture2f + *(tri + 0) * 2;
									tc2 = trace->model->surfmesh.data_texcoordtexture2f + *(tri + 1) * 2;
									tc3 = trace->model->surfmesh.data_texcoordtexture2f + *(tri + 2) * 2;

// find which of the 3 vertices has minimum and maximum u
									if (*tc1 < *tc2)
										if (*tc1 < *tc3)
										{
											min = tc1;pmin = v1;
											*tc2 > *tc3 ? (max = tc2, pmax = v2) : (max = tc3, pmax = v3);
										}
										else
										{
											min = tc3;pmin = v3;
											*tc1 > *tc2 ? (max = tc1, pmax = v1) : (max = tc2, pmax = v2);
										}
									else
										if (*tc2 < *tc3)
										{
											min = tc2;pmin = v2;
											*tc1 > *tc3 ? (max = tc1, pmax = v1) : (max = tc3, pmax = v3);
										}
										else
										{
											min = tc3;pmin = v3;
											*tc1 > *tc2 ? (max = tc1, pmax = v1) : (max = tc2, pmax = v2);
										}

// interpolate the traceEndposLocal u using the fraction of (traceEndposLocal - vert_min_u)'s projection onto
// (vert_max_u - vert_min_u)
									VectorSubtract (pmax, pmin, pminToPmax);
									VectorSubtract (traceEndposLocal, pmin, pminToEndpos);
									trace->texcoords[0] = DotProduct (pminToPmax, pminToEndpos) / DotProduct (pminToPmax, pminToPmax);
									trace->texcoords[0] = *min + (trace->texcoords[0] * (*max - *min));

									tc1 ++;tc2 ++;tc3 ++;

// ...then for v
									if (*tc1 < *tc2)
										if (*tc1 < *tc3)
										{
											min = tc1;pmin = v1;
											*tc2 > *tc3 ? (max = tc2, pmax = v2) : (max = tc3, pmax = v3);
										}
										else
										{
											min = tc3;pmin = v3;
											*tc1 > *tc2 ? (max = tc1, pmax = v1) : (max = tc2, pmax = v2);
										}
									else
										if (*tc2 < *tc3)
										{
											min = tc2;pmin = v2;
											*tc1 > *tc3 ? (max = tc1, pmax = v1) : (max = tc3, pmax = v3);
										}
										else
										{
											min = tc3;pmin = v3;
											*tc1 > *tc2 ? (max = tc1, pmax = v1) : (max = tc2, pmax = v2);
										}

									VectorSubtract (pmax, pmin, pminToPmax);
									VectorSubtract (traceEndposLocal, pmin, pminToEndpos);
									trace->texcoords[1] = DotProduct (pminToPmax, pminToEndpos) / DotProduct (pminToPmax, pminToPmax);
									trace->texcoords[1] = *min + (trace->texcoords[1] * (*max - *min));
								}
							}

							return;
						}
					}
				}
			}
		}
	}
};

trace_t CL_TraceBox(const vec3_t start, const vec3_t mins, const vec3_t maxs, const vec3_t end, int hitbmodels, int *hitent, int hitsupercontentsmask, qboolean hitplayers)
{
	int n;
	entity_render_t *ent;
	vec3_t tracemins, tracemaxs;
	trace_t cliptrace, trace;
	vec3_t origin;
	vec3_t starttransformed, endtransformed;
	vec3_t entmins, entmaxs;
	vec_t *playermins, *playermaxs;

	memset (&cliptrace, 0 , sizeof(trace_t));
	cliptrace.fraction = 1;
	cliptrace.realfraction = 1;

// VENOMUS - oh look a hack
	cliptrace.getModel = cliptrace.getTex = cliptrace.getUV = FALSE;
	if (hitbmodels & 2)
	{
		hitbmodels -= 2;
		cliptrace.getModel = TRUE;
	}
	if (hitbmodels & 4)
	{
		hitbmodels -= 4;
		cliptrace.getTex = TRUE;
	}
	if (hitbmodels & 8)
	{
		hitbmodels -= 8;
		cliptrace.getUV = TRUE;
	}
	(void *) cliptrace.model = (void *) cliptrace.tex = NULL;
	cliptrace.texcoords[0] = cliptrace.texcoords[1] = 0.0f;

	if (cl.worldmodel && cl.worldmodel->TraceBox)
		cl.worldmodel->TraceBox(cl.worldmodel, 0, &cliptrace, start, mins, maxs, end, hitsupercontentsmask);

// VENOMUS
//	if (hitent)
//		*hitent = 0;
	cliptrace.hitEnt = 0;

	if (hitbmodels && cl.num_brushmodel_entities)
	{
		tracemins[0] = min(start[0], end[0]) + mins[0];
		tracemaxs[0] = max(start[0], end[0]) + maxs[0];
		tracemins[1] = min(start[1], end[1]) + mins[1];
		tracemaxs[1] = max(start[1], end[1]) + maxs[1];
		tracemins[2] = min(start[2], end[2]) + mins[2];
		tracemaxs[2] = max(start[2], end[2]) + maxs[2];

		// look for embedded bmodels
		for (n = 0;n < cl.num_brushmodel_entities;n++)
		{
			ent = &cl.entities[cl.brushmodel_entities[n]].render;
			if (!BoxesOverlap(tracemins, tracemaxs, ent->mins, ent->maxs))
				continue;

			Matrix4x4_Transform(&ent->inversematrix, start, starttransformed);
			Matrix4x4_Transform(&ent->inversematrix, end, endtransformed);

			memset (&trace, 0 , sizeof(trace_t));
			trace.fraction = 1;
			trace.realfraction = 1;

// VENOMUS
			trace.getModel = cliptrace.getModel;
			trace.getTex = cliptrace.getTex;
			trace.getUV = cliptrace.getUV;
			(void *) trace.model = (void *) trace.tex = NULL;
			cliptrace.texcoords[0] = cliptrace.texcoords[1] = 0.0f;

			if (ent->model && ent->model->TraceBox)
				ent->model->TraceBox(ent->model, 0, &trace, starttransformed, mins, maxs, endtransformed, hitsupercontentsmask);

			// LordHavoc: take the 'best' answers from the new trace and combine with existing data
			if (trace.allsolid)
				cliptrace.allsolid = true;
			if (trace.startsolid)
			{
				cliptrace.startsolid = true;
				if (cliptrace.realfraction == 1)
// VENOMUS
//					if (hitent)
//						*hitent = cl.brushmodel_entities[n];
					cliptrace.hitEnt = cl.brushmodel_entities[n];
			}

			// don't set this except on the world, because it can easily confuse
			// monsters underwater if there's a bmodel involved in the trace
			// (inopen && inwater is how they check water visibility)
			//if (trace.inopen)
			//	cliptrace.inopen = true;
			if (trace.inwater)
				cliptrace.inwater = true;
			if (trace.realfraction < cliptrace.realfraction)
			{
				cliptrace.fraction = trace.fraction;
				cliptrace.realfraction = trace.realfraction;
				cliptrace.plane = trace.plane;
// VENOMUS
//				if (hitent)
//					*hitent = cl.brushmodel_entities[n];
				cliptrace.hitEnt = cl.brushmodel_entities[n];

				Matrix4x4_Transform3x3(&ent->matrix, trace.plane.normal, cliptrace.plane.normal);
				cliptrace.hitsupercontents = trace.hitsupercontents;
				cliptrace.hitq3surfaceflags = trace.hitq3surfaceflags;
				cliptrace.hittexture = trace.hittexture;

// VENOMUS
				cliptrace.model = trace.model;
				cliptrace.nodenum = trace.nodenum;
				cliptrace.tex = trace.tex;

			}
			cliptrace.startsupercontents |= trace.startsupercontents;
		}
	}
	if (hitplayers)
	{
		tracemins[0] = min(start[0], end[0]) + mins[0];
		tracemaxs[0] = max(start[0], end[0]) + maxs[0];
		tracemins[1] = min(start[1], end[1]) + mins[1];
		tracemaxs[1] = max(start[1], end[1]) + maxs[1];
		tracemins[2] = min(start[2], end[2]) + mins[2];
		tracemaxs[2] = max(start[2], end[2]) + maxs[2];

		for (n = 1;n < cl.maxclients+1;n++)
		{
			if (n != cl.playerentity)
			{
				ent = &cl.entities[n].render;
				// FIXME: crouch
				playermins = cl.playerstandmins;
				playermaxs = cl.playerstandmaxs;
				Matrix4x4_OriginFromMatrix(&ent->matrix, origin);
				VectorAdd(origin, playermins, entmins);
				VectorAdd(origin, playermaxs, entmaxs);
				if (!BoxesOverlap(tracemins, tracemaxs, entmins, entmaxs))
					continue;

				memset (&trace, 0 , sizeof(trace_t));
				trace.fraction = 1;
				trace.realfraction = 1;

// VENOMUS
				trace.getModel = cliptrace.getModel;
				trace.getTex = cliptrace.getTex;
				trace.getUV = cliptrace.getUV;
				(void *) trace.model = (void *) trace.tex = NULL;
				cliptrace.texcoords[0] = cliptrace.texcoords[1] = 0.0f;

				Matrix4x4_Transform(&ent->inversematrix, start, starttransformed);
				Matrix4x4_Transform(&ent->inversematrix, end, endtransformed);
				Collision_ClipTrace_Box(&trace, playermins, playermaxs, starttransformed, mins, maxs, endtransformed, hitsupercontentsmask, SUPERCONTENTS_BODY, 0, NULL);

				// LordHavoc: take the 'best' answers from the new trace and combine with existing data
				if (trace.allsolid)
					cliptrace.allsolid = true;
				if (trace.startsolid)
				{
					cliptrace.startsolid = true;
					if (cliptrace.realfraction == 1)
// VENOMUS
//						if (hitent)
//							*hitent = n;
						cliptrace.hitEnt = n;
				}
				// don't set this except on the world, because it can easily confuse
				// monsters underwater if there's a bmodel involved in the trace
				// (inopen && inwater is how they check water visibility)
				//if (trace.inopen)
				//	cliptrace.inopen = true;
				if (trace.inwater)
					cliptrace.inwater = true;
				if (trace.realfraction < cliptrace.realfraction)
				{
					cliptrace.fraction = trace.fraction;
					cliptrace.realfraction = trace.realfraction;
					cliptrace.plane = trace.plane;

// VENOMUS
//					if (hitent)
//						*hitent = n;
					cliptrace.hitEnt = n;

					Matrix4x4_Transform3x3(&ent->matrix, trace.plane.normal, cliptrace.plane.normal);
					cliptrace.hitsupercontents = trace.hitsupercontents;
					cliptrace.hitq3surfaceflags = trace.hitq3surfaceflags;
					cliptrace.hittexture = trace.hittexture;

// VENOMUS - we can return the model
					cliptrace.model = trace.model;
// VENOMUS - but model textures not supported yet!
					cliptrace.getTex = FALSE;
					cliptrace.getUV = FALSE;

				}
				cliptrace.startsupercontents |= trace.startsupercontents;
			}
		}
	}
	cliptrace.fraction = bound(0, cliptrace.fraction, 1);
	cliptrace.realfraction = bound(0, cliptrace.realfraction, 1);
	VectorLerp(start, cliptrace.fraction, end, cliptrace.endpos);

// VENOMUS - make it compatible with old hitent output
	if (hitent)
	{
		*hitent = cliptrace.hitEnt;
	}

	return cliptrace;
}

float CL_SelectTraceLine(const vec3_t start, const vec3_t end, vec3_t impact, vec3_t normal, int *hitent, entity_render_t *ignoreent, qboolean csqcents)
{
	float maxfrac, maxrealfrac;
	int n, entsnum;
	entity_t *entlist;
	unsigned char *entactivelist;
	entity_render_t *ent;
	float tracemins[3], tracemaxs[3];
	trace_t trace;
	float tempnormal[3], starttransformed[3], endtransformed[3];

	memset (&trace, 0 , sizeof(trace_t));
	trace.fraction = 1;
	trace.realfraction = 1;
	VectorCopy (end, trace.endpos);

	if (hitent)
		*hitent = 0;
	if (cl.worldmodel && cl.worldmodel->TraceBox)
		cl.worldmodel->TraceBox(cl.worldmodel, 0, &trace, start, vec3_origin, vec3_origin, end, SUPERCONTENTS_SOLID);

	if (normal)
		VectorCopy(trace.plane.normal, normal);
	maxfrac = trace.fraction;
	maxrealfrac = trace.realfraction;

	tracemins[0] = min(start[0], end[0]);
	tracemaxs[0] = max(start[0], end[0]);
	tracemins[1] = min(start[1], end[1]);
	tracemaxs[1] = max(start[1], end[1]);
	tracemins[2] = min(start[2], end[2]);
	tracemaxs[2] = max(start[2], end[2]);

	if(csqcents)
	{
		entlist = cl.csqcentities;
		entactivelist = cl.csqcentities_active;
		entsnum = cl.num_csqcentities;
	}
	else
	{
		entlist = cl.entities;
		entactivelist = cl.entities_active;
		entsnum = cl.num_entities;
	}

	// look for embedded bmodels
	for (n = 0;n < entsnum;n++)
	{
		if (!entactivelist[n])
			continue;
		ent = &entlist[n].render;
		if (!BoxesOverlap(ent->mins, ent->maxs, tracemins, tracemaxs))
			continue;
		if (!ent->model || !ent->model->TraceBox)
			continue;
		if ((ent->flags & RENDER_EXTERIORMODEL) && !chase_active.integer)
			continue;
		// if transparent and not selectable, skip entity
		if (!(entlist[n].state_current.effects & EF_SELECTABLE) && (ent->alpha < 1 || (ent->effects & (EF_ADDITIVE | EF_NODEPTHTEST))))
			continue;
		if (ent == ignoreent)
			continue;
		Matrix4x4_Transform(&ent->inversematrix, start, starttransformed);
		Matrix4x4_Transform(&ent->inversematrix, end, endtransformed);

		//if (ent->model && ent->model->TraceBox)
			ent->model->TraceBox(ent->model, ent->frameblend[0].frame, &trace, starttransformed, vec3_origin, vec3_origin, endtransformed, SUPERCONTENTS_SOLID);

		if (maxrealfrac > trace.realfraction)
		{
			if (hitent)
				*hitent = n;
			maxfrac = trace.fraction;
			maxrealfrac = trace.realfraction;
			if (normal)
			{
				VectorCopy(trace.plane.normal, tempnormal);
				Matrix4x4_Transform3x3(&ent->matrix, tempnormal, normal);
			}
		}
	}
	maxfrac = bound(0, maxfrac, 1);
	maxrealfrac = bound(0, maxrealfrac, 1);
	//if (maxfrac < 0 || maxfrac > 1) Con_Printf("fraction out of bounds %f %s:%d\n", maxfrac, __FILE__, __LINE__);
	if (impact)
		VectorLerp(start, maxfrac, end, impact);
	return maxfrac;
}

void CL_FindNonSolidLocation(const vec3_t in, vec3_t out, vec_t radius)
{
	// FIXME: check multiple brush models
	if (cl.worldmodel && cl.worldmodel->brush.FindNonSolidLocation)
		cl.worldmodel->brush.FindNonSolidLocation(cl.worldmodel, in, out, radius);
}

int CL_PointSuperContents(const vec3_t p)
{
	return CL_TraceBox(p, vec3_origin, vec3_origin, p, true, NULL, 0, false).startsupercontents;
}

