// VENOMUS
#include "r_modelreplacements.h"

// uncomment to dump model replacement textures
//#define DUMPTEX
#ifdef DUMPTEX
#include "image.h"
#endif

cvar_t r_model_replacements = {CVAR_SAVE, "r_model_replacements", "1", "replace certain models with custom special effects"};

static rtexture_t *laserboltorange;
static rtexture_t *laserboltred;
static rtexture_t *bubble;
static rtexturepool_t *modelreplacementstexturepool;

unsigned char shadebubble (float dx, float dy, vec3_t light);

void r_model_replacements_start (void)
{
	float dx, dy;
	int x, y, a;
	int resX, resY;
	int resXO2, resYO2;
	int coreResX, coreResY;
	int coreResXO2, coreResYO2;
	unsigned char pixels[64][64][4];
	vec3_t light;

	resX = 64;
	resY = 64;
	resXO2 = (int) (resX / 2.0f);
	resYO2 = (int) (resY / 2.0f);

	modelreplacementstexturepool = R_AllocTexturePool ();
	for (y = 0; y < resY; y ++)
	{
		dy = (y - (float) resYO2) / ((float) resYO2 + 1.0f);
		for (x = 0; x < resX; x ++)
		{
			dx = (x - (float) resXO2) / (float) resXO2;
			a = (int) (128.0f * 4.0f * (1 - (dx * dx + dy * dy)));
			pixels[y][x][0] = bound (0, a * 0.95f, 255);
			pixels[y][x][1] = bound (0, a * 0.42f, 255);
			pixels[y][x][2] = bound (0, a * 0.14f, 255);
			pixels[y][x][3] = 255;
		}
	}

	coreResX = (int) (resX / 1.3f);
	coreResY = (int) (resX / 1.3f);
	coreResXO2 = (int) (coreResX / 2.0f);
	coreResYO2 = (int) (coreResY / 2.0f);

	for (y = resYO2 - coreResYO2; y < resYO2 + coreResYO2; y ++)
	{
		dy = (y - (float) resYO2) / (float) coreResYO2;
		for (x = resXO2 - coreResXO2; x < resXO2 + coreResXO2; x ++)
		{
			dx = (x - (float) resXO2) / (float) coreResXO2;
			a = (int) (128.0f * 10.0f * (1 - (dx * dx + dy * dy)));
			pixels[y][x][0] = bound (0, pixels[y][x][0] + bound (0, a * 0.95f, 255), 255);
			pixels[y][x][1] = bound (0, pixels[y][x][1] + bound (0, a * 0.42f, 255), 255);
			pixels[y][x][2] = bound (0, pixels[y][x][2] + bound (0, a * 0.14f, 255), 255);
			pixels[y][x][3] = 255;
		}
	}

	laserboltorange = R_LoadTexture2D (modelreplacementstexturepool, "laserboltorange", resX, resY, &pixels[0][0][0], TEXTYPE_RGBA, TEXF_PRECACHE, NULL);
#ifdef DUMPTEX
	Image_WriteTGARGBA ("modelreplacementtex/laserboltorange.tga", resX, resY, &pixels[0][0][0]);
#endif
	for (y = 0; y < resY; y ++)
	{
		dy = (y - (float) resYO2) / ((float) resYO2 + 1.0f);
		for (x = 0; x < resX; x ++)
		{
			dx = (x - (float) resXO2) / (float) resXO2;
			a = (int) (128.0f * (1 - (dx * dx + dy * dy)));
			pixels[y][x][0] = bound (0, a * 1.0f, 255);
			pixels[y][x][1] = bound (0, a * 0.0f, 255);
			pixels[y][x][2] = bound (0, a * 0.0f, 255);
			pixels[y][x][3] = 255;
		}
	}

// this is a bit of a crap way to do the core, really need a formula that gives you a 'peak' rather than
// a quadratic buildup of alpha in the centre
	coreResX = (int) (resX / 2.0f);
	coreResY = (int) (resX / 2.0f);
	coreResXO2 = (int) (coreResX / 2.0f);
	coreResYO2 = (int) (coreResY / 2.0f);

	for (y = resYO2 - coreResYO2; y < resYO2 + coreResYO2; y ++)
	{
		dy = (y - (float) resYO2) / (float) coreResYO2;
		for (x = resXO2 - coreResXO2; x < resXO2 + coreResXO2; x ++)
		{
			dx = (x - (float) resXO2) / (float) coreResXO2;
			a = (int) (128.0f * 2.0f * (1 - (dx * dx + dy * dy)));
			pixels[y][x][0] = bound (0, pixels[y][x][0] + bound (0, a * 1.0f, 255), 255);
			pixels[y][x][1] = bound (0, pixels[y][x][1] + bound (0, a * 0.1f, 255), 255);
			pixels[y][x][2] = bound (0, pixels[y][x][2] + bound (0, a * 0.1f, 255), 255);
			pixels[y][x][3] = 255;
		}
	}

	laserboltred = R_LoadTexture2D (modelreplacementstexturepool, "laserboltred", resX, resY, &pixels[0][0][0], TEXTYPE_RGBA, TEXF_PRECACHE, NULL);
#ifdef DUMPTEX
	Image_WriteTGARGBA ("modelreplacementtex/laserboltred.tga", resX, resY, &pixels[0][0][0]);
#endif
// code from particle bubbles, they use a different texture pool
	light[0] = 1;
	light[1] = 1;
	light[2] = 1;
	VectorNormalize (light);
	for (y = 0; y < resY; y ++)
	{
		dy = (y - (float) resYO2) / ((float) resYO2);
		for (x = 0; x < resX; x ++)
		{
			dx = (x - (float) resXO2) / (float) resXO2;
			pixels[y][x][0] = pixels[y][x][1] = pixels[y][x][2] = shadebubble (dx, dy, light);
			pixels[y][x][3] = 255;
		}
	}

	bubble = R_LoadTexture2D (modelreplacementstexturepool, "bubble", resX, resY, &pixels[0][0][0], TEXTYPE_RGBA, TEXF_PRECACHE, NULL);
#ifdef DUMPTEX
	Image_WriteTGARGBA ("modelreplacementtex/bubble.tga", resX, resY, &pixels[0][0][0]);
#endif
};

void r_model_replacements_shutdown (void)
{
	modelreplacementstexturepool = NULL;
	laserboltorange = NULL;
	laserboltred = NULL;
};

void r_model_replacements_newmap (void)
{
};

void R_ModelReplacements_Init (void)
{
	Cvar_RegisterVariable(&r_model_replacements);
	R_RegisterModule("R_ModelReplacements", r_model_replacements_start, r_model_replacements_shutdown, r_model_replacements_newmap);
};

void R_Laser1_Draw_TransparentCallback (const entity_render_t *ent, const rtlight_t *rtlight, int numsurfaces, int *surfacelist)
{
	vec3_t look, left, up;
	float d;

	VectorSubtract (r_view.origin, ent->origin, look);
	VectorNormalize (look);
	left[0] = ent->matrix.m[0][0];
	left[1] = ent->matrix.m[1][0];
	left[2] = ent->matrix.m[2][0];
	CrossProduct (look, left, up);
	VectorNormalize (up);
	R_DrawSprite (GL_ONE, GL_ONE, laserboltorange, NULL, false, ent->origin, left, up, ent->scale * 12.0f, -ent->scale * 12.0f, -ent->scale * 2.0f, ent->scale * 2.0f, 1.0f, 1.0f, 1.0f, 1.0f);
	d = fabs (DotProduct (look, left)) * 2 - 1;
// at shallow angles add a circular, fully billboarded sprite to stop the cardboard cutout effect
// it doesn't look perfect and you can see 'popping' if you look carefully
// but 99.99% of players i don't think will notice
//	if (fabs (DotProduct (look, left)) > 0.9860132971832693404278880485936)
//	{
		R_DrawSprite (GL_ONE, GL_ONE, laserboltorange, NULL, false, ent->origin, r_view.right, r_view.up, ent->scale * 2.0f, -ent->scale * 2.0f, -ent->scale * 2.0f, ent->scale * 2.0f, d, d, d, 1.0f);
//	}
};

void R_Laser1_Draw (entity_render_t *ent)
{
	if (r_model_replacements.value > 0)
	{
		R_MeshQueue_AddTransparent (ent->origin, R_Laser1_Draw_TransparentCallback, ent, 0, r_shadow_rtlight);
	}
	else
	{
		R_Q1BSP_Draw (ent);
	}
};

void R_Laser2_Draw_TransparentCallback (const entity_render_t *ent, const rtlight_t *rtlight, int numsurfaces, int *surfacelist)
{
	vec3_t look, left, up;
	float d;

	VectorSubtract (r_view.origin, ent->origin, look);
	VectorNormalize (look);
	left[0] = ent->matrix.m[0][0];
	left[1] = ent->matrix.m[1][0];
	left[2] = ent->matrix.m[2][0];
	CrossProduct (look, left, up);
	VectorNormalize (up);
	R_DrawSprite (GL_ONE, GL_ONE, laserboltred, NULL, false, ent->origin, left, up, ent->scale * 12.0f, -ent->scale * 12.0f, -ent->scale * 2.0f, ent->scale * 2.0f, 1.0f, 1.0f, 1.0f, 1.0f);
	d = fabs (DotProduct (look, left)) * 2 - 1;
//	if (fabs (DotProduct (look, left)) > 0.9860132971832693404278880485936)
//	{
		R_DrawSprite (GL_ONE, GL_ONE, laserboltred, NULL, false, ent->origin, r_view.right, r_view.up, ent->scale * 2.0f, -ent->scale * 2.0f, -ent->scale * 2.0f, ent->scale * 2.0f, d, d, d, 1.0f);
//	}
};

void R_Laser2_Draw (entity_render_t *ent)
{
	if (r_model_replacements.value > 0)
	{
		R_MeshQueue_AddTransparent (ent->origin, R_Laser2_Draw_TransparentCallback, ent, 0, r_shadow_rtlight);
	}
	else
	{
		R_Q1BSP_Draw (ent);
	}
};

void R_Model_Sprite_Draw (entity_render_t *ent);

void R_Bubble_Draw_TransparentCallback (const entity_render_t *ent, const rtlight_t *rtlight, int numsurfaces, int *surfacelist)
{
	float rad;
// some way of blending radius between frames would be nice
// but air_bubbles sprites don't even seem to have working sprite interpolation
	rad = ent->frame == 0 ? 4.0f : 2.0f;
// should actual be a random grey between 0.25 and 0.5
	R_DrawSprite (GL_ONE, GL_ONE, bubble, NULL, false, ent->origin, r_view.right, r_view.up, ent->scale * rad, -ent->scale * rad, -ent->scale * rad, ent->scale * rad, 0.4, 0.4, 0.4, 1.0f);
};

void R_Bubble_Draw (entity_render_t *ent)
{
	if (r_model_replacements.value > 0)
	{
		R_MeshQueue_AddTransparent (ent->origin, R_Bubble_Draw_TransparentCallback, ent, 0, r_shadow_rtlight);
	}
	else
	{
		R_Model_Sprite_Draw (ent);
	}
};

void R_Replaced_CompileShadowVolume (entity_render_t *ent, vec3_t relativelightorigin, float lightradius, int numsurfaces, const int *surfacelist)
{
	if (r_model_replacements.value <= 0)
	{
		R_Q1BSP_CompileShadowVolume (ent, relativelightorigin, lightradius, numsurfaces, surfacelist);
	}
};

void R_Replaced_DrawShadowVolume (entity_render_t *ent, vec3_t relativelightorigin, float lightradius, int modelnumsurfaces, const int *modelsurfacelist, const vec3_t lightmins, const vec3_t lightmaxs)
{
	if (r_model_replacements.value <= 0)
	{
		R_Q1BSP_DrawShadowVolume (ent, relativelightorigin, lightradius, modelnumsurfaces, modelsurfacelist, lightmins, lightmaxs);
	}
};

void R_Replaced_DrawLight (entity_render_t *ent, int numsurfaces, const int *surfacelist)
{
	if (r_model_replacements.value <= 0)
	{
		R_Q1BSP_DrawLight (ent, numsurfaces, surfacelist);
	}
};
// VENOMUS END